import asyncio
import aiohttp

try:
    import ujson as json
except ImportError:
    import json

from mydm.type import ApiResponse
from mydm.event import Event
from mydm.utils import EchoHandler
from mydm.interactor import InteractorApi, InteractorEvent
from mydm.exceptions import WebSocketNotConnected

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from mydm.type import HANDLER


__all__ = [
    'InteractorApiHttp',
    'InteractorWebSocket',
]


class InteractorApiHttp(InteractorApi):
    def __init__(
        self, url: str,
        *,
        access_token: str = '',
        timeout: float = 10
    ):
        super().__init__(url, access_token=access_token, timeout=timeout)
        self.session = aiohttp.ClientSession(json_serialize=json.dumps)

    async def call(self, action: str, **params) -> 'ApiResponse':
        data = {
            'action': action,
            'params': params,
        }
        async with self.session.post(self.url, data=data, timeout=self._timeout) as msg:
            resp = await msg.json()
        resp = ApiResponse(resp)
        return resp


class InteractorWebSocket(InteractorApi, InteractorEvent):
    def __init__(
        self, url: str,
        *,
        access_token: str = '',
        timeout: float = 10
    ):
        super().__init__(url, access_token=access_token, timeout=timeout)
        self.session = aiohttp.ClientSession(json_serialize=json.dumps)
        self.handlers: list['HANDLER'] = []
        self.conn = False

    async def connect(self, event_loop: 'AbstractEventLoop'):
        self.ws_session = await self.session.ws_connect(self.url)
        self.conn = True
        await self.receive(event_loop)

    async def close(self):
        if not self.conn:
            raise WebSocketNotConnected
        await self.ws_session.close()
        self.conn = False

    async def call(self, action: str, **params) -> 'ApiResponse':
        if not self.conn:
            raise WebSocketNotConnected

        echo = EchoHandler.next()
        EchoHandler.add_future(echo)
        data = {
            'action': action,
            'params': params,
            'echo': echo,
        }
        await self.ws_session.send_json(data)

        resp = await EchoHandler.wait(echo, self._timeout)
        resp = ApiResponse(resp)
        return resp

    async def _receiver(self):
        while True:
            data = await self.ws_session.receive_json()
            if data.get('echo') is None:
                # 正常上报
                event = Event(data)
                asyncio.gather(*[handler(event) for handler in self.handlers])
            else:
                # echo上报
                EchoHandler.set(data)



__version__ = (0, 0, 4)


from mydm import bot
from mydm import type
from mydm import event
from mydm import utils
from mydm import cqcode
from mydm import message
from mydm import interactor
from mydm import exceptions



from typing import Literal, Optional
from mydm.exceptions import DataFormatError

__all__ = [
    'MessageSegment',
    'MessageSegmentReceive',
    'MessageSegmentSend',
    'Message',
]


class MessageSegment(dict):
    def __init__(self, data: dict):
        """从dict构建MessageSegment"""
        super().__init__(data)
        try:
            _ = self.type
        except KeyError as e:
            raise DataFormatError(e)

    def __add__(self, __other) -> 'Message':
        """合并Message"""
        return Message(self, __other)

    def __eq__(self, __value: 'MessageSegment') -> bool:
        return self.type == __value.type and self.data == __value.data

    @property
    def type(self) -> Literal['text', 'face', 'record', 'video', 'at', 'share', 'image', 'reply', 'xml', 'json', 'redbag', 'forward', 'share', 'music', 'poke', 'gift', 'node', 'cardimage', 'tts']:
        """消息段类型"""
        return self['type']

    @property
    def data(self) -> Optional[dict]:
        """数据"""
        return self.get('data')


class MessageSegmentReceive(MessageSegment):
    def __init__(self, data: dict):
        """从dict构建MessageSegmentReceive"""
        super().__init__(data)
        try:
            _ = self.type
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def type(self) -> Literal['text', 'face', 'record', 'video', 'at', 'share', 'image', 'reply', 'xml', 'json', 'redbag', 'forward']:
        """消息段类型"""
        return self['type']


class MessageSegmentSend(MessageSegment):
    def __init__(self, data: dict):
        """
        从dict构建MessageSegmentSend
        """
        super().__init__(data)
        try:
            _ = self.type
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def type(self) -> Literal['text', 'face', 'record', 'video', 'at', 'share', 'image', 'reply', 'xml', 'json', 'share', 'music', 'poke', 'gift', 'node', 'cardimage', 'tts']:
        """消息段类型"""
        return self['type']


class Message(list[MessageSegment]):
    def __init__(self, *data):
        """从dict构建Message"""
        if not data:
            raise DataFormatError('data is empty')
        super().__init__()
        for segment in data:
            if isinstance(segment, Message):
                self.extend(segment)
            elif isinstance(segment, MessageSegment):
                self.append(segment)
            elif isinstance(segment, (list, tuple)):
                self.extend(MessageSegment(s) for s in segment)
            elif isinstance(segment, dict):
                self.append(MessageSegment(segment))
            else:
                raise DataFormatError(
                    f'{segment} is not a MessageSegment or Message'
                )

    def __add__(self, __other):
        """合并Message"""
        return Message(self, __other)



from abc import ABC, abstractmethod

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from mydm.type import HANDLER, ApiResponse

__all__ = [
    'InteractorApi',
    'InteractorEvent',
]


class InteractorApi(ABC):
    def __init__(
        self, url: str,
        *,
        access_token: str = '',
        timeout: float = 10
    ):
        self.url = url
        self.access_token = access_token
        self._timeout = timeout

    @abstractmethod
    async def call(self, action: str, **params) -> 'ApiResponse':
        """请求并获取返回值"""


class InteractorEvent(ABC):
    def __init__(
        self, url: str,
        *,
        access_token: str = '',
    ):
        self.url = url
        self.access_token = access_token
        self.handlers: list[HANDLER] = []

    @abstractmethod
    async def _receiver(self):
        """接收Event"""

    async def receive(self, event_loop: 'AbstractEventLoop'):
        event_loop.create_task(self._receiver())



import sys
import asyncio
from typing import Any, Callable, Coroutine, TYPE_CHECKING

from mydm.event import *
from mydm.exceptions import ApiCallTimeout

if TYPE_CHECKING:
    from asyncio import Future

__all__ = [
    'EchoHandler',
    'EventHandler',
    'event_handler_sort',
]


class EchoHandler:
    _seq = 0
    _echos: dict[int, 'Future'] = {}

    @classmethod
    def next(cls) -> int:
        """获取下一个echo号"""
        seq = cls._seq
        cls._seq = (cls._seq + 1) % sys.maxsize
        return seq

    @classmethod
    def set(cls, data: dict):
        """设置echo"""
        echo = data.get('echo')
        if isinstance(echo, int):
            future = cls._echos[echo]
            future.set_result(data)

    @classmethod
    def add_future(cls, echo: int):
        future = asyncio.get_event_loop().create_future()
        cls._echos[echo] = future

    @classmethod
    async def wait(cls, echo: int, timeout: float) -> dict:
        """等待echo被set，返回整条消息"""
        future = cls._echos[echo]
        try:
            return await asyncio.wait_for(future, timeout)
        except asyncio.TimeoutError:
            raise ApiCallTimeout
        finally:
            cls._echos.pop(echo)


class EventHandler:
    def __init__(
        self, handler: Callable[['Event'], Any],
        priority: int = 0,
        condition: dict | None = None
    ):
        self.handler = handler
        self.priority = priority
        self.condition = condition or {}

    def assert_condition(self, condition: dict) -> bool:
        """判断condition"""
        if not self.condition:
            return True
        try:
            for key, value in self.condition.items():
                if isinstance(value, list):
                    for v in value:
                        if condition[key] == v:
                            return True
                else:
                    if condition[key] == value:
                        return True
        except KeyError:
            return False
        return False

    async def __call__(self, raw_event: 'Event'):
        post_type = self.condition.get('post_type')
        match post_type:
            case 'message' | 'message_sent':
                event = EventMessage(raw_event)
            case 'request':
                event = EventRequest(raw_event)
            case 'notice':
                event = EventNotice(raw_event)
            case 'meta_event':
                event = EventMeta(raw_event)
            case _:
                event = raw_event
        handler = self.handler(event)
        if isinstance(handler, Coroutine):
            return await handler
        else:
            return handler


def event_handler_sort(handlers: list[EventHandler]) -> list[EventHandler]:
    """事件处理器排序"""
    def key(handler: EventHandler):
        value = handler.priority
        return (True, 0) if value == 0 else (value < 0, abs(value))
    sorted_handlers = [k for k in sorted(handlers, key=key)]
    return sorted_handlers



from typing import Literal, Optional
from mydm.message import MessageSegmentSend

__all__ = [
    'text',
    'face',
    'at',
    'reply',
]


# TODO 下面的CQ码构造并没有写全，参考https://docs.go-cqhttp.org/cqcode/

def build(type: str, **data):
    """
    构建MessageSegmentSend
    """
    for k in list(data.keys()):
        if data[k] is None:
            data.pop(k)
        elif isinstance(data[k], int):
            data[k] = str(data[k])
    return MessageSegmentSend({
        'type': type,
        'data': data
    })


def text(text: str):
    """纯文本
    - `text`: 文本
    """
    return build('text', text=text)


def face(id: int):
    """QQ 表情
    - `id`: QQ 表情 ID，参考https://github.com/kyubotics/coolq-http-api/wiki/%E8%A1%A8%E6%83%85-CQ-%E7%A0%81-ID-%E8%A1%A8
    """
    return build('face', id=id)


def at(qq: int | Literal['all'], name: Optional[str] = None):
    """@某人
    - `qq`: @的 QQ 号, all 表示全体成员
    - `name`: 当在群中找不到此QQ号的名称时才会生效
    """
    return build('at', qq=qq, name=name)


def reply(
    id: int, *,
    text: Optional[str] = None,
    qq: Optional[int] = None,
    time: Optional[int] = None,
    seq: Optional[int] = None,
):
    """回复
    - `id`: 回复时所引用的消息id, 必须为本群消息
    - `text`: 自定义回复的信息
    - `qq`: 自定义回复时的自定义QQ, 如果使用自定义信息必须指定
    - `time`: 自定义回复时的时间, 格式为Unix时间
    - `seq`: 起始消息序号, 可通过 `get_msg` 获得
    """
    return build('reply', id=id, text=text, qq=qq, time=time, seq=seq)



from typing import Any, Callable, Literal, TYPE_CHECKING, Union

from mydm.exceptions import ApiCallFailed, DataFormatError

if TYPE_CHECKING:
    from mydm.event import *

try:
    import ujson as json
except ImportError:
    import json

__all__ = [
    'ApiResponse',
    'Sender',
    'SenderTemp',
    'SenderGroup',
    'POST_TYPE'
]


POST_TYPE = Literal[
    'message', 'message_sent', 'request', 'notice', 'meta_event'
]
HANDLER = Callable[
    [Union['Event', 'EventMessage', 'EventRequest', 'EventNotice', 'EventMeta']], Any
]


class ApiResponse(dict):
    def __init__(
        self, resp: dict | str
    ):
        super().__init__()
        if isinstance(resp, str):
            resp = json.loads(resp)
        if isinstance(resp, str):
            # Pylance无法识别，只好这样
            raise
        try:
            retcode = resp['retcode']
            if retcode == 0:
                # api调用成功
                self.update(resp['data'])
            elif retcode == 1:
                # api已提交 async 处理
                pass
            else:
                # 操作失败
                raise ApiCallFailed(resp['msg'], resp['wording'])
        except KeyError:
            pass


class Sender(dict):
    def __init__(self, data: dict):
        """
        从dict构建Sender
        """
        super().__init__(data)
        try:
            _ = self.user_id, self.nickname, self.sex, self.age
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def user_id(self) -> int:
        """QQ号"""
        return self['user_id']

    @property
    def nickname(self) -> str:
        """昵称"""
        return self['nickname']

    @property
    def sex(self) -> Literal['male', 'female', 'unknown']:
        """性别"""
        return self['sex']

    @property
    def age(self) -> int:
        """年龄"""
        return self['age']


class SenderTemp(Sender):
    def __init__(self, data: dict):
        """
        从dict构建SenderTemp
        """
        super().__init__(data)
        try:
            _ = self.group_id
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def group_id(self) -> int | None:
        """临时群消息来源群号"""
        return self['group_id']


class SenderGroup(Sender):
    def __init__(self, data: dict):
        """
        从dict构建SenderGroup
        """
        super().__init__(data)
        try:
            _ = self.card, self.area, self.level, self.role, self.title
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def card(self) -> str:
        """群名片/备注"""
        return self['card']

    @property
    def area(self) -> str:
        """地区"""
        return self['area']

    @property
    def level(self) -> str:
        """等级"""
        return self['level']

    @property
    def role(self) -> Literal['owner', 'admin', 'member']:
        """角色"""
        return self['role']

    @property
    def title(self) -> str:
        """头衔"""
        return self['title']



from typing import Callable, Optional, TYPE_CHECKING, Union

from mydm.utils import event_handler_sort, EventHandler
from mydm.exceptions import StopEventProcessing

if TYPE_CHECKING:
    from mydm.type import POST_TYPE, HANDLER
    from mydm.event import Event

__all__ = [
    'Bot',
]


class Bot:
    def __init__(self):
        self.handlers: list['EventHandler'] = []

    async def __call__(self, event: 'Event') -> None:
        """
        根据优先级依次执行处理函数
        """
        sorted_handlers = event_handler_sort(self.handlers)
        try:
            for handler in sorted_handlers:
                if handler.assert_condition(event):
                    await handler(event)
        except StopEventProcessing:
            return

    def on(
        self,
        *,
        priority: int = 0,
        post_type: Optional[Union['POST_TYPE', list['POST_TYPE']]] = None,
        **condition: str | list
    ) -> Callable:
        """注册事件处理函数
        - `priority`: 优先级：正数中，越小优先级越高；0为无优先级，其次执行；负数最后执行；如果优先级相同，则按照添加顺序执行
        - `post_type`: 上报类型，默认为全部类型
        - `**condition`: 事件满足条件才会触发
        """
        def decorator(func: 'HANDLER'):
            if func.__code__.co_argcount != 1:
                raise TypeError('Handler function must have only one arg')
            if post_type is not None:
                condition.update({'post_type': post_type})
            self.handlers.append(
                EventHandler(func, priority=priority, condition=condition)
            )
            return func

        return decorator



from typing import Literal, TYPE_CHECKING

from mydm.type import Sender
from mydm.message import Message
from mydm.exceptions import DataFormatError

if TYPE_CHECKING:
    from mydm.type import POST_TYPE


__all__ = [
    'Event',
    'EventMessage',
    'EventRequest',
    'EventNotice',
    'EventMeta',
]


class Event(dict):
    """事件"""

    def __init__(self, data: dict):
        """
        从dict构建Event
        """
        super().__init__(data)
        try:
            _ = self.time, self.self_id, self.post_type
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def time(self) -> int:
        """事件发生的unix时间戳"""
        return self['time']

    @property
    def self_id(self) -> int:
        """收到事件的机器人的 QQ 号"""
        return self['self_id']

    @property
    def post_type(self) -> 'POST_TYPE':
        """上报类型"""
        return self['post_type']


class EventMessage(Event):
    """
    消息事件
    """

    def __init__(self, data: dict):
        """
        从dict构建EventMessage
        """
        super().__init__(data)
        try:
            self._message = Message(self['message'])
            self._sender = Sender(self['sender'])
            _ = self.post_type, self.message_type, self.sub_type, self.message_id, self.user_id, self.raw_message, self.font
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def post_type(self) -> Literal['message', 'message_sent']:
        """上报类型"""
        return self['post_type']

    @property
    def message_type(self) -> Literal['private', 'group']:
        """消息类型"""
        return self['message_type']

    @property
    def sub_type(self) -> Literal['friend', 'normal', 'anonymous', 'group_self', 'group', 'notice']:
        """消息子类型"""
        return self['sub_type']

    @property
    def message_id(self) -> int:
        """消息ID"""
        return self['message_id']

    @property
    def user_id(self) -> int:
        """消息发送者的 QQ 号"""
        return self['user_id']

    @property
    def message(self) -> 'Message':
        """消息内容"""
        return self._message

    @property
    def raw_message(self) -> str:
        """原始消息内容"""
        return self['raw_message']

    @property
    def font(self) -> int:
        """字体"""
        return self['font']

    @property
    def sender(self) -> 'Sender':
        """消息发送者"""
        return self._sender


class EventRequest(Event):
    def __init__(self, data: dict):
        """
        从dict构建EventRequest
        """
        super().__init__(data)
        try:
            _ = self.post_type, self.request_type
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def post_type(self) -> Literal['request']:
        """上报类型"""
        return self['post_type']

    @property
    def request_type(self) -> Literal['friend', 'group']:
        """请求类型"""
        return self['request_type']


class EventNotice(Event):
    def __init__(self, data: dict):
        """
        从dict构建EventNotice
        """
        super().__init__(data)
        try:
            _ = self.post_type, self.notice_type
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def post_type(self) -> Literal['request']:
        """上报类型"""
        return self['post_type']

    @property
    def notice_type(self) -> Literal['group_upload', 'group_admin', 'group_decrease', 'group_increase', 'group_ban', 'friend_add', 'group_recall', 'friend_recall', 'group_card', 'offline_file', 'client_status', 'essence', 'notify']:
        """请求类型"""
        return self['notice_type']


class EventMeta(Event):
    def __init__(self, data: dict):
        """
        从dict构建EventMeta
        """
        super().__init__(data)
        try:
            _ = self.meta_event_type
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def meta_event_type(self) -> Literal['lifecycle', 'heartbeat']:
        """元事件类型"""
        return self['meta_event_type']



__all__ = [
    'ApiCallFailed',
    'ApiCallTimeout',
    'WebSocketNotConnected',
    'DataFormatError',
    'StopEventProcessing',
]


class ApiCallFailed(Exception):
    """Api调用失败"""

    def __init__(self, msg: str, wording: str):
        super().__init__(msg, wording)


class ApiCallTimeout(BaseException):
    """Api调用超时"""


class WebSocketNotConnected(Exception):
    """WebSocket未连接"""


class DataFormatError(TypeError):
    """数据格式错误"""


class StopEventProcessing(Exception):
    """截停事件处理"""
